import dataclasses
import random

from PowerfulDiceRoller.models.errors import NegativeDiceFaces


@dataclasses.dataclass
class Dice:
    __slots__ = ("_face", "__result")

    def __init__(self, face: int):
        if not isinstance(face, int):
            raise TypeError(f"Face type must be {type(int)}, not {type(face)}")
        elif not face > 0:
            raise NegativeDiceFaces(f"Face value must be >= 1, not {face}")
        self._face: int = face
        self.__result: int = random.randint(1, self._face)

    @property
    def face(self) -> int:
        return self._face

    @property
    def result(self) -> int:
        return self.__result

    def set_result(self, num: int) -> None:
        self.__result = num

    def __int__(self) -> int:
        return self.result

    def __str__(self) -> str:
        return str(self.result)

    def __repr__(self) -> str:
        return f"Dice(face={self._face}, __result={self.result})"

    def __add__(self, other) -> int:  # +
        if isinstance(other, type(self)):
            return self.result + other.result
        return self.result + other

    def __sub__(self, other):  # -
        if isinstance(other, type(self)):
            return self.result - other.result
        return self.result - other

    def __mul__(self, other):  # *
        if isinstance(other, type(self)):
            return self.result * other.result
        return self.result * other

    def __truediv__(self, other):  # /
        if isinstance(other, type(self)):
            return self.result / other.result
        return self.result / other

    def __floordiv__(self, other):  # //
        if isinstance(other, type(self)):
            return self.result // other.result
        return self.result // other

    def __radd__(self, other) -> int:
        return other + self.result

    def __rsub__(self, other):
        return other - self.result

    def __rmul__(self, other):
        return other * self.result

    def __rtruediv__(self, other):
        return other / self.result

    def __rfloordiv__(self, other):
        return other // self.result

    def __eq__(self, other):  # ==
        if isinstance(other, type(self)):
            return self.result == other.result

        return self.result == other

    def __lt__(self, other):  # <=
        if isinstance(other, type(self)):
            return self.result < other.result
        return self.result < other


from PowerfulDiceRoller.models.Dice import Dice
from PowerfulDiceRoller.models.DiceThrown import DiceThrown
from PowerfulDiceRoller.models.LarkThrown import LarkThrown
from PowerfulDiceRoller.models.errors import DiceError, NotFoundMethod, ParseError, NegativeDiceFaces

__all__ = (
    "Dice",
    "DiceThrown",
    "LarkThrown",
    "DiceError",
    "NotFoundMethod",
    "ParseError",
    "NegativeDiceFaces"
)



from PowerfulDiceRoller.models.errors.DiceError import DiceError
from PowerfulDiceRoller.models.errors.NotFoundMethod import NotFoundMethod
from PowerfulDiceRoller.models.errors.ParseError import ParseError
from PowerfulDiceRoller.models.errors.NegativeDiceFaces import NegativeDiceFaces


__all__ = (
    "DiceError",
    "NotFoundMethod",
    "ParseError",
    "NegativeDiceFaces"
)


#!/usr/bin/env python3
import pathlib
import re
import sys

from setuptools import find_packages, setup

WORK_DIR = pathlib.Path(__file__).parent
MINIMAL_PY_VERSION = (3, 7)

if sys.version_info < MINIMAL_PY_VERSION:
    raise RuntimeError('PowerfulDiceRoller works only with Python {}+'.format('.'.join(map(str, MINIMAL_PY_VERSION))))


def get_version() -> str:
    """
    Read version

    :return: str
    """
    txt = (WORK_DIR / 'PowerfulDiceRoller' / '__init__.py').read_text('utf-8')
    try:
        return re.findall(r"^__version__ = '([^']+)'\r?$", txt, re.M)[0]
    except IndexError:
        raise RuntimeError('Unable to determine version.')

this_directory = pathlib.Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='PowerfulDiceRoller',
    version=get_version(),
    packages=find_packages(exclude=('tests', 'tests.*')),
    url='https://github.com/M9SCO/PowerfulDiceRoller',
    license='MIT',
    author='Chernov M_9SCO Grigorii',
    author_email='chr.grigorii@gmail.com',
    python_requires='>=3.7',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
    install_requires=[
        'lark>=1.1.2'
    ],
    include_package_data=False,
)


import typing

from PowerfulDiceRoller.models.Dice import Dice
from PowerfulDiceRoller.models.errors.DiceError import DiceError


class DiceThrown:
    __slots__ = ("_times",
                 "_face",
                 "_amount_function",
                 "_amount_rate",
                 "_result",
                 "_amount",
                 "_dices",
                 )

    def __init__(self,
                 times: int,
                 faces: int,
                 dropout_function: typing.Callable = None,
                 dropout_rate: int = None) -> None:
        self._times = times
        self._face = faces
        self._amount_function = dropout_function
        self._amount_rate = dropout_rate

        self._amount = None

        self._dices: typing.List[Dice] = [Dice(face=faces) for _ in range(self._times)]

        if any((dropout_function, dropout_rate)) and not all((dropout_rate, dropout_function)):
            raise DiceError

    def to_str(self, view_strike=False, startswith_strike='<strike>', endswith_strike='</strike>') -> str:
        values = list(map(str, self.dices))
        if view_strike:
            amount = self.amount.copy()
            for i, dice in enumerate(self.dices):
                if dice in amount:
                    amount.remove(dice)
                else:
                    values[i] = f"{startswith_strike}{dice.result}{endswith_strike}"
        return f"[{'+'.join(values)}]"

    def __repr__(self) -> str:
        return f"DiceThrown(throw={self.times}, face={self.face}, amount_function = {self._amount_function}, " \
               f"amount_rate = {self._amount_rate}, result = {self.total})"

    def _get_amount(self) -> typing.Optional[typing.List[Dice]]:
        if self._amount_function is None and self._amount_rate is None:
            return self._dices
        if any((self._amount_function, self._amount_rate)) and not all((self._amount_function, self._amount_rate)):
            raise DiceError("Unspecified retain formula or count for save")
        elif self._amount_rate > self._times:
            return self._dices
        dices = self._dices.copy()
        results = []
        for _ in range(self._amount_rate):
            exclude = self._amount_function(dices)
            dices.remove(exclude)
            results.append(exclude)
        return results

    @property
    def times(self) -> int:
        return self._times

    @property
    def face(self) -> int:
        return self._face

    @property
    def amount(self) -> typing.Optional[typing.List[Dice]]:
        if self._amount_function is None and self._amount_rate is None:
            return self.dices
        if self._amount is None:
            self._amount = self._get_amount()
        return self._amount

    @property
    def total(self) -> int:
        if self.amount:
            return sum(self.amount)
        return sum(self._dices)

    @property
    def dices(self) -> typing.List[Dice]:
        return self._dices
